private variable MODULE_NAME = "mpfp";
prepend_to_slang_load_path (".");
set_import_module_path (".:"+get_import_module_path ());
require (MODULE_NAME);
require ("rand");

private variable Failed = 0;
private define failed ()
{
   variable args = __pop_list (_NARGS);
   () = fprintf (stderr, __push_list (args));
   () = fputs ("\n", stderr);
   Failed++;
}

private define is_equal (a, b)
{
   variable diff = abs(a-b);
   variable abs_a = abs(a);
   variable abs_b = abs(b);
   variable res = all ((isnan(a)==isnan(b))
		       and (isnan(a) or (a==b)
			    or (diff < 1e-15)
			    or ((abs_a > abs_b) and (diff < abs_a*1e-10))
			    or (abs(a-b) < abs_b*1e-10)));
   if (res) return res;
   %print (abs(a-b));
   return 0;
   return ((isnan(a)==isnan(b))
	   && (isnan(a) || (a==b) || (abs(a-b)<1e-15)));
   % Note; the a==b is for Inf.
}

private define mpfr_test_op (z, x, y, op)
{
   variable e1, e2, e3;
   variable left, right;
   (left, right) = mpfi_interval (x);
   e2 = (left + right)/2;

   (left, right) = mpfi_interval (y);
   e1 = (left + right)/2;

   e3 = (@op)(e2, e1);
   (left, right) = mpfi_interval (z);
   ifnot (left <= e3 <= right)
     failed ("op %S. e2=%S, e1=%S, e3=%S", op, e2, e1, e3);
}

private define mpfi_test (x, y)
{
   mpfr_test_op (x+y, x, y, &_op_plus);
   mpfr_test_op (x-y, x, y, &_op_minus);
   mpfr_test_op (x*y, x, y, &_op_times);
   mpfr_test_op (x/y, x, y, &_op_divide);
}

private define test_1 ()
{
   % This test was adapted from the one distributed with the mpfi
   % source code.
   variable n = 1000;
   variable s = 1;
   loop (n)
     {
	variable x, y, e, f;
	e = mpfr(rand_uniform ());
	e = 1/e;
	f = mpfr(rand_uniform());
	x = mpfi(f);
	x = x/f;

	e = mpfr(rand_uniform());
	e = 1/e;
	f = mpfr(rand_uniform());
	y = mpfi(f);
	y = y/f;

	y *= s;
	s = -s;

	mpfi_test (x, y);
     }
}

private define test_cmp ()
{
   variable vals = {0, 1, -1, 0.0, 1.0, -1.0, -_Inf, _Inf, _NaN};
   variable ops = {&_op_eqs, &_op_neqs, &_op_lt, &_op_le, &_op_gt, &_op_ge};
   variable x, y, op;
   foreach x (vals)
     {
	variable xfi = mpfi(x);
	variable xfr = mpfr(x);
	foreach y (vals)
	  {
	     variable yfi = mpfi(y);
	     variable yfr = mpfr(y);
	     foreach op (ops)
	       {
		  variable c0, c1;
		  c0 = (@op)(x, y);

		  c1 = (@op)(x, yfi);
		  if (c0 != c1)
		    failed ("%S %S %S produced %S, expected %S",
			    x, op, yfi, c1, c0);

		  c1 = (@op)(xfr, yfi);
		  if (c0 != c1)
		    failed ("%Sm %S %S produced %S, expected %S",
			    xfr, op, yfi, c1, c0);

		  c1 = (@op)(xfi, y);
		  if (c0 != c1)
		    failed ("%S %S %S produced %S, expected %S",
			    xfi, op, y, c1, c0);

		  c1 = (@op)(xfi, yfr);
		  if (c0 != c1)
		    failed ("%S %S %Sm produced %S, expected %S",
			    xfi, op, yfr, c1, c0);

		  c1 = (@op)(xfi, yfi);
		  if (c0 != c1)
		    failed ("%S %S %S produced %S, expected %S",
			    xfi, op, yfi, c1, c0);
	       }
	  }
     }
}

private define test_new_mpfi ()
{
   variable types =
     {
	Char_Type, Short_Type, Int_Type, Long_Type,
	UChar_Type, UShort_Type, UInt_Type, ULong_Type,
	Float_Type, Double_Type, MPFR_Type
     };

   variable x = [1:10], a, b, z, t1, t2;
   foreach t1 (types)
     {
	a = typecast (x, t1);
	z = mpfi (a);
	foreach t2 (types)
	  {
	     b = typecast (1, t2);
	     z = mpfi (b, b);
	     z = mpfi (a, b);
	     z = mpfi (b, a);
	     b = typecast (x+1, t2);
	     z = mpfi (a, b);
	     variable left, right;
	     (left, right) = mpfi_interval (z);
	     ifnot (is_equal (left, a))
	       failed ("left end of interval not what expected");
	     ifnot (is_equal (right, b))
	       failed ("right end of interval not what expected");
	  }
     }
}

define slsh_main ()
{
   ()= mpfi(2.0);
   () = fprintf (stdout, "Testing mpfi..."); () = fflush(stdout);
   variable prec = 53;
   mpfr_set_default_prec (prec);

   test_new_mpfi ();
   test_1 ();
   test_cmp ();

   if (Failed == 0)
     () = fprintf (stdout, "Ok\n");
   else
     () = fprintf (stderr, "Failed with %d errors\n", Failed);
}

